import { toast } from "react-toastify";
import { PizzaOrder } from "./types";
import { getBaseUrl, getToken } from "./Utils";

/**
 * Wrapper pro volání API, u kterých chceme automaticky zobrazit toaster s chybou ze serveru.
 * 
 * @param apiFunction volaná API funkce
 */
export function errorHandler<T>(apiFunction: () => Promise<T>): Promise<T> {
    return new Promise<T>((resolve, reject) => {
        apiFunction().then((result) => {
            resolve(result);
        }).catch(e => {
            toast.error(e.message, { theme: "colored" });
        });
    });
}

async function request<TResponse>(
    url: string,
    config: RequestInit = {}
): Promise<TResponse> {
    config.headers = config?.headers ? new Headers(config.headers) : new Headers();
    config.headers.set("Authorization", `Bearer ${getToken()}`);
    try {
        const response = await fetch(getBaseUrl() + url, config);
        if (!response.ok) {
            const json = await response.json();
            // Vyhodíme samotnou hlášku z odpovědi, odchytí si jí errorHandler
            throw new Error(json.error);
        }
        return response.json() as TResponse;
    } catch (e) {
        return Promise.reject(e);
    }
}

const api = {
    get: <TResponse>(url: string) => request<TResponse>(url),
    post: <TBody extends BodyInit, TResponse>(url: string, body: TBody) => request<TResponse>(url, { method: 'POST', body, headers: { 'Content-Type': 'application/json' } }),
}

export const getQrUrl = (login: string) => {
    return `${getBaseUrl()}/api/qr?login=${login}`;
}

export const getData = async (dayIndex?: number) => {
    let url = '/api/data';
    if (dayIndex != null) {
        url += '?dayIndex=' + dayIndex;
    }
    return await api.get<any>(url);
}
export const createPizzaDay = async () => {
    return await api.post<any, any>('/api/createPizzaDay', undefined);
}

export const deletePizzaDay = async () => {
    return await api.post<any, any>('/api/deletePizzaDay', undefined);
}

export const lockPizzaDay = async () => {
    return await api.post<any, any>('/api/lockPizzaDay', undefined);
}

export const unlockPizzaDay = async () => {
    return await api.post<any, any>('/api/unlockPizzaDay', undefined);
}

export const finishOrder = async () => {
    return await api.post<any, any>('/api/finishOrder', undefined);
}

export const finishDelivery = async (bankAccount?: string, bankAccountHolder?: string) => {
    return await api.post<any, any>('/api/finishDelivery', JSON.stringify({ bankAccount, bankAccountHolder }));
}

export const addChoice = async (locationIndex: number, foodIndex?: number, dayIndex?: number) => {
    return await api.post<any, any>('/api/addChoice', JSON.stringify({ locationIndex, foodIndex, dayIndex }));
}

export const removeChoices = async (locationIndex: number, dayIndex?: number) => {
    return await api.post<any, any>('/api/removeChoices', JSON.stringify({ locationIndex, dayIndex }));
}

export const removeChoice = async (locationIndex: number, foodIndex: number, dayIndex?: number) => {
    return await api.post<any, any>('/api/removeChoice', JSON.stringify({ locationIndex, foodIndex, dayIndex }));
}

export const addPizza = async (pizzaIndex: number, pizzaSizeIndex: number) => {
    return await api.post<any, any>('/api/addPizza', JSON.stringify({ pizzaIndex, pizzaSizeIndex }));
}

export const removePizza = async (pizzaOrder: PizzaOrder) => {
    return await api.post<any, any>('/api/removePizza', JSON.stringify({ pizzaOrder }));
}

export const updatePizzaDayNote = async (note?: string) => {
    return await api.post<any, any>('/api/updatePizzaDayNote', JSON.stringify({ note }));
}

export const login = async (login?: string) => {
    return await api.post<any, any>('/api/login', JSON.stringify({ login }));
}

export const changeDepartureTime = async (login: string, time: string, dayIndex?: number) => {
    return await api.post<any, any>('/api/changeDepartureTime', JSON.stringify({ login, time, dayIndex }));
}