import express from "express";
import { Server } from "socket.io";
import bodyParser from "body-parser";
import cors from 'cors';
import { addChoice, getData, getDateForWeekIndex, getToday, removeChoice, removeChoices, updateDepartureTime } from "./service";
import { addPizzaOrder, createPizzaDay, deletePizzaDay, finishPizzaDelivery, finishPizzaOrder, getPizzaList, lockPizzaDay, removePizzaOrder, unlockPizzaDay, updatePizzaDayNote } from "./pizza";
import dotenv from 'dotenv';
import path from 'path';
import { getQr } from "./qr";
import { generateToken, getLogin, getTrusted, verify } from "./auth";
import { InsufficientPermissions, getDayOfWeekIndex } from "./utils";

const ENVIRONMENT = process.env.NODE_ENV || 'production';
dotenv.config({ path: path.resolve(__dirname, `./.env.${ENVIRONMENT}`) });

// Validace nastavení JWT tokenu - nemá bez něj smysl vůbec povolit server spustit
if (!process.env.JWT_SECRET) {
    throw Error("Není vyplněna proměnná prostředí JWT_SECRET");
}

const app = express();
const server = require("http").createServer(app);
const io = new Server(server, {
    cors: {
        origin: "*",
    },
});

// Body-parser middleware for parsing JSON
app.use(bodyParser.json());
// app.use(express.json());

app.use(cors({
    origin: '*'
}));

// app.use((req, res, next) => {
//     console.log("--- Request ---")
//     console.log(req.url);
//     console.log(req.baseUrl);
//     console.log(req.originalUrl);
//     console.log(req.path);
//     next();
// });

app.use(express.static('public'))

const parseToken = (req: any) => {
    if (req?.headers?.authorization) {
        return req.headers.authorization.split(' ')[1];
    }
}

/**
 * Ověří a vrátí index dne v týdnu z požadavku, za předpokladu, že byl předán, a je zároveň
 * roven nebo vyšší indexu dnešního dne.
 * 
 * @param req request
 * @returns index dne v týdnu
 */
const parseValidateFutureDayIndex = (req: any) => {
    if (req.body.dayIndex == null) {
        throw Error(`Nebyl předán index dne v týdnu.`);
    }
    const todayDayIndex = getDayOfWeekIndex(getToday());
    const dayIndex = parseInt(req.body.dayIndex);
    if (isNaN(dayIndex)) {
        throw Error(`Neplatný index dne v týdnu: ${req.body.dayIndex}`);
    }
    if (dayIndex < todayDayIndex) {
        throw Error(`Předaný index dne v týdnu (${dayIndex}) nesmí být nižší než dnešní den (${todayDayIndex})`);
    }
    return dayIndex;
}

// ----------- Metody nevyžadující token --------------

app.get("/api/whoami", (req, res) => {
    res.send(req.header('remote-user'));
})

app.post("/api/login", (req, res) => {
    // Autentizace pomocí trusted headers
    const remoteUser = req.header('remote-user');
    const remoteName = req.header('remote-name');
    if (remoteUser && remoteUser.length > 0 && remoteName && remoteName.length > 0) {
        res.status(200).json(generateToken(Buffer.from(remoteName, 'latin1').toString(), true));
        return;
    }
    // Klasická autentizace loginem
    if (!req.body?.login || req.body.login.trim().length === 0) {
        throw Error("Nebyl předán login");
    }
    // TODO zavést podmínky pro délku loginu (min i max)
    res.status(200).json(generateToken(req.body.login, false));
});

// TODO dočasné řešení - QR se zobrazuje přes <img>, nemáme sem jak dostat token
app.get("/api/qr", (req, res) => {
    // const login = getLogin(parseToken(req));
    if (!req.query?.login) {
        throw Error("Nebyl předán login");
    }
    const img = getQr(req.query.login as string);
    res.writeHead(200, {
        'Content-Type': 'image/png',
        'Content-Length': img.length
    });
    res.end(img);
});

// ----------------------------------------------------

/** Middleware ověřující JWT token */
app.use((req, res, next) => {
    const userHeader = req.header('remote-user');
    const nameHeader = req.header('remote-name');
    const emailHeader = req.header('remote-email');
    if (userHeader !== undefined && nameHeader !== undefined) {
        const remoteName = Buffer.from(nameHeader, 'latin1').toString();
        console.log("Tvuj username, name a email: %s, %s, %s.", userHeader, remoteName, emailHeader);
    }
    if (!req.headers.authorization) {
        return res.status(401).json({ error: 'Nebyl předán autentizační token' });
    }
    const token = req.headers.authorization.split(' ')[1];
    if (!verify(token)) {
        return res.status(403).json({ error: 'Neplatný autentizační token' });
    }
    next();
});

/** Vrátí data pro aktuální den. */
app.get("/api/data", async (req, res) => {
    let date = undefined;
    if (req.query.dayIndex != null && typeof req.query.dayIndex === 'string') {
        const index = parseInt(req.query.dayIndex);
        if (!isNaN(index)) {
            date = getDateForWeekIndex(parseInt(req.query.dayIndex));
        }
    }
    res.status(200).json(await getData(date));
});

/** Založí pizza day pro aktuální den, za předpokladu že dosud neexistuje. */
app.post("/api/createPizzaDay", async (req, res) => {
    const login = getLogin(parseToken(req));
    const data = await createPizzaDay(login);
    res.status(200).json(data);
    io.emit("message", data);
});

/** Smaže pizza day pro aktuální den, za předpokladu že existuje. */
app.post("/api/deletePizzaDay", async (req, res) => {
    const login = getLogin(parseToken(req));
    const data = await deletePizzaDay(login);
    io.emit("message", data);
});

app.post("/api/addPizza", async (req, res) => {
    const login = getLogin(parseToken(req));
    if (isNaN(req.body?.pizzaIndex)) {
        throw Error("Nebyl předán index pizzy");
    }
    const pizzaIndex = req.body.pizzaIndex;
    if (isNaN(req.body?.pizzaSizeIndex)) {
        throw Error("Nebyl předán index velikosti pizzy");
    }
    const pizzaSizeIndex = req.body.pizzaSizeIndex;
    let pizzy = await getPizzaList();
    if (!pizzy) {
        throw Error("Selhalo získání seznamu dostupných pizz.");
    }
    if (!pizzy[pizzaIndex]) {
        throw Error("Neplatný index pizzy: " + pizzaIndex);
    }
    if (!pizzy[pizzaIndex].sizes[pizzaSizeIndex]) {
        throw Error("Neplatný index velikosti pizzy: " + pizzaSizeIndex);
    }
    const data = await addPizzaOrder(login, pizzy[pizzaIndex], pizzy[pizzaIndex].sizes[pizzaSizeIndex]);
    io.emit("message", data);
    res.status(200).json({});
});

app.post("/api/removePizza", async (req, res) => {
    const login = getLogin(parseToken(req));
    if (!req.body?.pizzaOrder) {
        throw Error("Nebyla předána objednávka");
    }
    const data = await removePizzaOrder(login, req.body?.pizzaOrder);
    io.emit("message", data);
    res.status(200).json({});
});

app.post("/api/lockPizzaDay", async (req, res) => {
    const login = getLogin(parseToken(req));
    const data = await lockPizzaDay(login);
    io.emit("message", data);
    res.status(200).json({});
});

app.post("/api/unlockPizzaDay", async (req, res) => {
    const login = getLogin(parseToken(req));
    const data = await unlockPizzaDay(login);
    io.emit("message", data);
    res.status(200).json({});
});

app.post("/api/finishOrder", async (req, res) => {
    const login = getLogin(parseToken(req));
    const data = await finishPizzaOrder(login);
    io.emit("message", data);
    res.status(200).json({});
});

app.post("/api/finishDelivery", async (req, res) => {
    const login = getLogin(parseToken(req));
    const data = await finishPizzaDelivery(login, req.body.bankAccount, req.body.bankAccountHolder);
    io.emit("message", data);
    res.status(200).json({});
});

app.post("/api/addChoice", async (req, res, next) => {
    const login = getLogin(parseToken(req));
    const trusted = getTrusted(parseToken(req));
    if (req.body.locationIndex > -1) {
        let date = undefined;
        if (req.body.dayIndex != null) {
            let dayIndex;
            try {
                dayIndex = parseValidateFutureDayIndex(req);
            } catch (e: any) {
                return res.status(400).json({ error: e.message });
            }
            date = getDateForWeekIndex(dayIndex);
        }
        try {
            const data = await addChoice(login, trusted, req.body.locationIndex, req.body.foodIndex, date);
            io.emit("message", data);
            return res.status(200).json(data);
        } catch (e: any) { next(e) }
    }
    return res.status(400); // TODO přidat popis chyby
});

app.post("/api/removeChoices", async (req, res, next) => {
    const login = getLogin(parseToken(req));
    const trusted = getTrusted(parseToken(req));
    let date = undefined;
    if (req.body.dayIndex != null) {
        let dayIndex;
        try {
            dayIndex = parseValidateFutureDayIndex(req);
        } catch (e: any) {
            return res.status(400).json({ error: e.message });
        }
        date = getDateForWeekIndex(dayIndex);
    }
    try {
        const data = await removeChoices(login, trusted, req.body.locationIndex, date);
        io.emit("message", data);
        res.status(200).json(data);
    } catch (e: any) { next(e) }
});

app.post("/api/removeChoice", async (req, res, next) => {
    const login = getLogin(parseToken(req));
    const trusted = getTrusted(parseToken(req));
    let date = undefined;
    if (req.body.dayIndex != null) {
        let dayIndex;
        try {
            dayIndex = parseValidateFutureDayIndex(req);
        } catch (e: any) {
            return res.status(400).json({ error: e.message });
        }
        date = getDateForWeekIndex(dayIndex);
    }
    try {
        const data = await removeChoice(login, trusted, req.body.locationIndex, req.body.foodIndex, date);
        io.emit("message", data);
        res.status(200).json(data);
    } catch (e: any) { next(e) }
});

app.post("/api/updatePizzaDayNote", async (req, res) => {
    const login = getLogin(parseToken(req));
    if (req.body.note && req.body.note.length > 100) {
        throw Error("Poznámka může mít maximálně 100 znaků");
    }
    const data = await updatePizzaDayNote(login, req.body.note);
    io.emit("message", data);
    res.status(200).json(data);
});

app.post("/api/changeDepartureTime", async (req, res, next) => {
    const login = getLogin(parseToken(req));
    let date = undefined;
    if (req.body.dayIndex != null) {
        let dayIndex;
        try {
            dayIndex = parseValidateFutureDayIndex(req);
        } catch (e: any) {
            return res.status(400).json({ error: e.message });
        }
        date = getDateForWeekIndex(dayIndex);
    }
    try {
        const data = await updateDepartureTime(login, req.body?.time, date);
        io.emit("message", data);
        res.status(200).json(data);
    } catch (e: any) { next(e) }
});

// Middleware pro zpracování chyb
app.use((err: any, req: any, res: any, next: any) => {
    if (err instanceof InsufficientPermissions) {
        res.status(403).send({ error: err.message })
    } else {
        res.status(500).send({ error: err.message })
    }
    next();
});

io.on("connection", (socket) => {
    console.log(`New client connected: ${socket.id}`);

    socket.on("message", (message) => {
        io.emit("message", message);
    });

    socket.on("disconnect", () => {
        console.log(`Client disconnected: ${socket.id}`);
    });
});

const PORT = process.env.PORT || 3001;
const HOST = process.env.HOST || '0.0.0.0';

server.listen(PORT, () => {
    console.log(`Server listening on ${HOST}, port ${PORT}`);
});

// Umožníme vypnutí serveru přes SIGINT, jinak Docker čeká než ho sestřelí
process.on('SIGINT', function () {
    console.log("\nSIGINT (Ctrl-C), vypínám server");
    process.exit(0);
});